#include "RedditContentProvider.h"
#include "util/SimpleThread.h"
#include "util/ThreadWorker.h"
#include "API/RedditAPI.h"

namespace ui
{
    void RedditContentProvider::GetContentAsync(std::function<void(std::vector<RedditContentListItemRef>)> onFinsihed, std::function<void()> onFailed)
    {
        if (LoadingContent())
        {
            return;
        }

        if (!m_HasMoreContent)
        {
            return;
        }

        struct ThreadStorage
        {
            std::vector<RedditContentListItemRef> content;
            bool error;
        };
        std::lock_guard<std::mutex> lock(m_ThreadMutex);
        m_Thread = new util::SimpleThread(
        [this](util::ThreadWorker* worker)
            {
                if (worker->IsCancelled())
                    return (void*)nullptr;

                //ensure user has been loaded before loading posts.
                api::RedditAPI::Get()->GetCurrentUser();

                auto storage = new ThreadStorage();
                storage->content = GetContent(worker, &storage->error);
                return (void*)storage;
            },
            [this, onFinsihed, onFailed](void* user_data, bool cancelled, util::SimpleThread*)
            {
                auto storage = static_cast<ThreadStorage*>(user_data);
                if (!cancelled)
                {
                    if (storage->error)
                    {
                        onFailed();
                    }
                    else
                    {
                        onFinsihed(storage->content);
                    }

                    delete storage;
                }

                std::lock_guard<std::mutex> lock(m_ThreadMutex);
                m_Thread = nullptr;
            });
    }

    std::vector<RedditContentListItemRef> RedditContentProvider::GetContent(util::ThreadWorker* worker, bool* error)
    {
        if (worker->IsCancelled())
            return std::vector<RedditContentListItemRef>();

        std::string timestamp = m_LastTimeStamp;
        std::vector<RedditContentListItemRef> newContent = GetContentInternal(timestamp, worker);

        if (worker->IsCancelled())
            return std::vector<RedditContentListItemRef>();

        if (!newContent.empty())
        {
            m_LastTimeStamp = timestamp;
        }

        if (m_HasMoreContent && newContent.empty())
        {
            *error = true;
            return std::vector<RedditContentListItemRef>();
        }

        if (worker->IsCancelled())
            return std::vector<RedditContentListItemRef>();

        if (m_LastTimeStamp.empty())
        {
            m_HasMoreContent = false;
        }

        if (worker->IsCancelled())
            return std::vector<RedditContentListItemRef>();

        return newContent;
    }

    void RedditContentProvider::Reset()
    {
        m_LastTimeStamp.clear();
        m_HasMoreContent = true;

        std::lock_guard<std::mutex> lock(m_ThreadMutex);
        if (m_Thread)
        {
            m_Thread->Cancel();
        }
    }

    bool RedditContentProvider::LoadingContent()
    {
        std::lock_guard<std::mutex> lock(m_ThreadMutex);
        return m_Thread != nullptr;
    }
}
