#include "HttpClient.h"
#include "Helpers.h"
#include <curl/curl.h>
#include <errno.h>
#include <AppSettings.h>

namespace util
{
    static std::map<CURLcode, std::string> s_ErrorCodeMap =
    {
        { CURLE_OK, "CURLE_OK"},
        { CURLE_UNSUPPORTED_PROTOCOL, "CURLE_UNSUPPORTED_PROTOCOL - 1"},
        { CURLE_FAILED_INIT, "CURLE_FAILED_INIT - 2"},
        { CURLE_URL_MALFORMAT, "CURLE_URL_MALFORMAT - 3"},
        { CURLE_NOT_BUILT_IN, "CURLE_NOT_BUILT_IN - 4"},
        { CURLE_COULDNT_RESOLVE_PROXY, "CURLE_COULDNT_RESOLVE_PROXY - 5"},
        { CURLE_COULDNT_RESOLVE_HOST, "CURLE_COULDNT_RESOLVE_HOST - 6"},
        { CURLE_COULDNT_CONNECT, "CURLE_COULDNT_CONNECT - 7"},
        { CURLE_WEIRD_SERVER_REPLY, "CURLE_WEIRD_SERVER_REPLY - 8"},
        { CURLE_REMOTE_ACCESS_DENIED, "CURLE_REMOTE_ACCESS_DENIED - 9 a service was denied by the server due to lack of access - when login fails this is not returned."},
        { CURLE_FTP_ACCEPT_FAILED, "CURLE_FTP_ACCEPT_FAILED - 10"},
        { CURLE_FTP_WEIRD_PASS_REPLY, "CURLE_FTP_WEIRD_PASS_REPLY - 11"},
        { CURLE_FTP_ACCEPT_TIMEOUT, "CURLE_FTP_ACCEPT_TIMEOUT - 12 - timeout occurred accepting server"},
        { CURLE_FTP_WEIRD_PASV_REPLY, "CURLE_FTP_WEIRD_PASV_REPLY - 13"},
        { CURLE_FTP_WEIRD_227_FORMAT, "CURLE_FTP_WEIRD_227_FORMAT - 14"},
        { CURLE_FTP_CANT_GET_HOST, "CURLE_FTP_CANT_GET_HOST - 15"},
        { CURLE_HTTP2, "CURLE_HTTP2 - 16 - A problem in the http2 framing layer."},
        { CURLE_FTP_COULDNT_SET_TYPE, "CURLE_FTP_COULDNT_SET_TYPE - 17"},
        { CURLE_PARTIAL_FILE, "CURLE_PARTIAL_FILE - 18"},
        { CURLE_FTP_COULDNT_RETR_FILE, "CURLE_FTP_COULDNT_RETR_FILE - 19"},
        { CURLE_OBSOLETE20, "CURLE_OBSOLETE20 - 20 - NOT USED"},
        { CURLE_QUOTE_ERROR, "CURLE_QUOTE_ERROR - 21 - quote command failure"},
        { CURLE_HTTP_RETURNED_ERROR, "CURLE_HTTP_RETURNED_ERROR - 22"},
        { CURLE_WRITE_ERROR, "CURLE_WRITE_ERROR - 23"},
        { CURLE_OBSOLETE24, "CURLE_OBSOLETE24 - 24 - NOT USED"},
        { CURLE_UPLOAD_FAILED, "CURLE_UPLOAD_FAILED - 25 - failed upload \"command\""},
        { CURLE_READ_ERROR, "CURLE_READ_ERROR - 26 - couldn't open/read from file"},
        { CURLE_OUT_OF_MEMORY, "CURLE_OUT_OF_MEMORY - 27"},
        { CURLE_OPERATION_TIMEDOUT, "CURLE_OPERATION_TIMEDOUT - 28 - the timeout time was reached"},
        { CURLE_OBSOLETE29, "CURLE_OBSOLETE29 - 29 - NOT USED"},
        { CURLE_FTP_PORT_FAILED, "CURLE_FTP_PORT_FAILED - 30 - FTP PORT operation failed"},
        { CURLE_FTP_COULDNT_USE_REST, "CURLE_FTP_COULDNT_USE_REST - 31 - the REST command failed"},
        { CURLE_OBSOLETE32, "CURLE_OBSOLETE32 - 32 - NOT USED"},
        { CURLE_RANGE_ERROR, "CURLE_RANGE_ERROR - 33 - RANGE \"command\" didn't work"},
        { CURLE_HTTP_POST_ERROR, "CURLE_HTTP_POST_ERROR - 34"},
        { CURLE_SSL_CONNECT_ERROR, "CURLE_SSL_CONNECT_ERROR - 35 - wrong when connecting with SSL"},
        { CURLE_BAD_DOWNLOAD_RESUME, "CURLE_BAD_DOWNLOAD_RESUME - 36 - couldn't resume download"},
        { CURLE_FILE_COULDNT_READ_FILE, "CURLE_FILE_COULDNT_READ_FILE - 37"},
        { CURLE_LDAP_CANNOT_BIND, "CURLE_LDAP_CANNOT_BIND - 38"},
        { CURLE_LDAP_SEARCH_FAILED, "CURLE_LDAP_SEARCH_FAILED - 39"},
        { CURLE_OBSOLETE40, "CURLE_OBSOLETE40 - 40 - NOT USED"},
        { CURLE_FUNCTION_NOT_FOUND, "CURLE_FUNCTION_NOT_FOUND - 41 - NOT USED starting with 7.53.0"},
        { CURLE_ABORTED_BY_CALLBACK, "CURLE_ABORTED_BY_CALLBACK - 42"},
        { CURLE_BAD_FUNCTION_ARGUMENT, "CURLE_BAD_FUNCTION_ARGUMENT - 43"},
        { CURLE_OBSOLETE44, "CURLE_OBSOLETE44 - 44 - NOT USED"},
        { CURLE_INTERFACE_FAILED, "CURLE_INTERFACE_FAILED - 45 - CURLOPT_INTERFACE failed"},
        { CURLE_OBSOLETE46, "CURLE_OBSOLETE46 - 46 - NOT USED"},
        { CURLE_TOO_MANY_REDIRECTS, "CURLE_TOO_MANY_REDIRECTS - 47 - catch endless re-direct loops"},
        { CURLE_UNKNOWN_OPTION, "CURLE_UNKNOWN_OPTION - 48 - User specified an unknown option"},
        { CURLE_TELNET_OPTION_SYNTAX, "CURLE_TELNET_OPTION_SYNTAX - 49 - Malformed telnet option"},
        { CURLE_OBSOLETE50, "CURLE_OBSOLETE50 - 50 - NOT USED"},
        { CURLE_OBSOLETE51, "CURLE_OBSOLETE51 - 51 - NOT USED"},
        { CURLE_GOT_NOTHING, "CURLE_GOT_NOTHING - 52 - when this is a specific error"},
        { CURLE_SSL_ENGINE_NOTFOUND, "CURLE_SSL_ENGINE_NOTFOUND - 53 - SSL crypto engine not found"},
        { CURLE_SSL_ENGINE_SETFAILED, "CURLE_SSL_ENGINE_SETFAILED - 54 - can not set SSL crypto engine as default"},
        { CURLE_SEND_ERROR, "CURLE_SEND_ERROR - 55 - failed sending network data"},
        { CURLE_RECV_ERROR, "CURLE_RECV_ERROR - 56 - failure in receiving network data"},
        { CURLE_OBSOLETE57, "CURLE_OBSOLETE57 - 57 - NOT IN USE"},
        { CURLE_SSL_CERTPROBLEM, "CURLE_SSL_CERTPROBLEM - 58 - problem with the local certificate"},
        { CURLE_SSL_CIPHER, "CURLE_SSL_CIPHER - 59 - couldn't use specified cipher"},
        { CURLE_PEER_FAILED_VERIFICATION, "CURLE_PEER_FAILED_VERIFICATION - 60 - peer's certificate or fingerprint wasn't verified fine"},
        { CURLE_BAD_CONTENT_ENCODING, "CURLE_BAD_CONTENT_ENCODING - 61 - Unrecognized/bad encoding"},
        { CURLE_LDAP_INVALID_URL, "CURLE_LDAP_INVALID_URL - 62 - Invalid LDAP URL"},
        { CURLE_FILESIZE_EXCEEDED, "CURLE_FILESIZE_EXCEEDED - 63 - Maximum file size exceeded"},
        { CURLE_USE_SSL_FAILED, "CURLE_USE_SSL_FAILED - 64 - Requested FTP SSL level failed"},
        { CURLE_SEND_FAIL_REWIND, "CURLE_SEND_FAIL_REWIND - 65 - Sending the data requires a rewind that failed"},
        { CURLE_SSL_ENGINE_INITFAILED, "CURLE_SSL_ENGINE_INITFAILED - 66 - failed to initialise ENGINE"},
        { CURLE_LOGIN_DENIED, "CURLE_LOGIN_DENIED - 67 - user, password or similar was not accepted and we failed to login"},
        { CURLE_TFTP_NOTFOUND, "CURLE_TFTP_NOTFOUND - 68 - file not found on server"},
        { CURLE_TFTP_PERM, "CURLE_TFTP_PERM - 69 - permission problem on server"},
        { CURLE_REMOTE_DISK_FULL, "CURLE_REMOTE_DISK_FULL - 70 - out of disk space on server"},
        { CURLE_TFTP_ILLEGAL, "CURLE_TFTP_ILLEGAL - 71 - Illegal TFTP operation"},
        { CURLE_TFTP_UNKNOWNID, "CURLE_TFTP_UNKNOWNID - 72 - Unknown transfer ID"},
        { CURLE_REMOTE_FILE_EXISTS, "CURLE_REMOTE_FILE_EXISTS - 73 - File already exists"},
        { CURLE_TFTP_NOSUCHUSER, "CURLE_TFTP_NOSUCHUSER - 74 - No such user"},
        { CURLE_CONV_FAILED, "CURLE_CONV_FAILED - 75 - conversion failed"},
        { CURLE_CONV_REQD, "CURLE_CONV_REQD - 76 - caller must register conversion"},
        { CURLE_SSL_CACERT_BADFILE, "CURLE_SSL_CACERT_BADFILE - 77 - could not load CACERT file, missing or wrong format"},
        { CURLE_REMOTE_FILE_NOT_FOUND, "CURLE_REMOTE_FILE_NOT_FOUND - 78 - remote file not found"},
        { CURLE_SSH, "CURLE_SSH - 79 - error from the SSH layer, somewhat generic so the error message will be of interest when this has happened"},
        { CURLE_SSL_SHUTDOWN_FAILED, "CURLE_SSL_SHUTDOWN_FAILED - 80 - Failed to shut down the SSL connection"},
        { CURLE_AGAIN, "CURLE_AGAIN - 81 - socket is not ready for send/recv, wait till it's ready and try again (Added in 7.18.2)"},
        { CURLE_SSL_CRL_BADFILE, "CURLE_SSL_CRL_BADFILE - 82 - could not load CRL file, missing or wrong format (Added in 7.19.0)"},
        { CURLE_SSL_ISSUER_ERROR, "CURLE_SSL_ISSUER_ERROR - 83 - Issuer check failed.  (Added in 7.19.0)"},
        { CURLE_FTP_PRET_FAILED, "CURLE_FTP_PRET_FAILED - 84 - a PRET command failed"},
        { CURLE_RTSP_CSEQ_ERROR, "CURLE_RTSP_CSEQ_ERROR - 85 - mismatch of RTSP CSeq numbers"},
        { CURLE_RTSP_SESSION_ERROR, "CURLE_RTSP_SESSION_ERROR - 86 - mismatch of RTSP Session Ids"},
        { CURLE_FTP_BAD_FILE_LIST, "CURLE_FTP_BAD_FILE_LIST - 87 - unable to parse FTP file list"},
        { CURLE_CHUNK_FAILED, "CURLE_CHUNK_FAILED - 88 - chunk callback reported error"},
        { CURLE_NO_CONNECTION_AVAILABLE, "CURLE_NO_CONNECTION_AVAILABLE - 89 - No connection available, the session will be queued"},
        { CURLE_SSL_PINNEDPUBKEYNOTMATCH, "CURLE_SSL_PINNEDPUBKEYNOTMATCH - 90 - specified pinned public key did not match"},
        { CURLE_SSL_INVALIDCERTSTATUS, "CURLE_SSL_INVALIDCERTSTATUS - 91 - invalid certificate status"},
        { CURLE_HTTP2_STREAM, "CURLE_HTTP2_STREAM - 92 - stream error in HTTP/2 framing laye"},
        { CURLE_RECURSIVE_API_CALL, "CURLE_RECURSIVE_API_CALL - 93 - an api function was called from inside a callback"},
        { CURLE_AUTH_ERROR, "CURLE_AUTH_ERROR - 94 - an authentication function returned an error"},
        { CURLE_HTTP3, "CURLE_HTTP3 - 95 - An HTTP/3 layer problem"},
        { CURLE_QUIC_CONNECT_ERROR, "CURLE_QUIC_CONNECT_ERROR - 96 - QUIC connection error"},
        { CURLE_PROXY, "CURLE_PROXY - 97 - proxy handshake error"},
        { CURL_LAST,  "CURL_LAST  - never use!"}
    };

    struct ResponseCallbackArgs
    {
        std::string& response;
    };

    size_t GetResponseCallback(void* ptr, size_t size, size_t nmemb, void* userdata)
    {
        ResponseCallbackArgs* args = static_cast<ResponseCallbackArgs*>(userdata);
        args->response += (const char*) ptr;
        return size * nmemb;
    }

    std::string HttpClient::Get(std::string url, std::vector<std::string> headerValues) const
    {
        std::string response;
        ResponseCallbackArgs args{response};

        CURL* curl = curl_easy_init();

        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, GetResponseCallback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*) &args);
        curl_easy_setopt(curl, CURLOPT_USERAGENT, "Headlines/1.0");
        struct curl_slist* hs = NULL;
        for (const std::string& header : headerValues)
        {
            hs = curl_slist_append(hs, header.c_str());
        }

        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, hs);

        CURLcode res = curl_easy_perform(curl);
        curl_slist_free_all(hs);
        curl_easy_cleanup(curl);

        if (AppSettings::Get()->GetBool("print_api_responses", false))
        {
            std::cout << "HttpClient::Get: " << url << std::endl;
            std::cout << "Response: " << util::Helpers::UnescapeHtml(response) << std::endl << std::endl;
        }

        if (res)
        {
            if (s_ErrorCodeMap.find(res) != s_ErrorCodeMap.end())
            {
                g_warning("Httplient::Get error: %s - %s", url.c_str(), s_ErrorCodeMap[res].c_str());
            } else
            {
                g_warning("Httplient::Get error: %s - %d", url.c_str(), (int) res);
            }
            return "";
        }

        return response;
    }

    size_t ImageDownloadedCallback(void* ptr, size_t size, size_t nmemb, void* userdata)
    {
        FILE* stream = (FILE*) userdata;
        if (!stream)
        {
            return 0;
        }

        size_t written = fwrite((FILE*) ptr, size, nmemb, stream);
        return written;
    }

    bool HttpClient::DownloadImage(std::string url, std::string outFileName, std::vector<std::string> headerValues) const
    {
        std::string sanitisedUrl = util::Helpers::UnescapeHtml(url);

        errno = 0;
        FILE* file = fopen(outFileName.c_str(), "wb");
        if (!file)
        {
            g_warning("failed to open file for writing: %s with error %d", outFileName.c_str(), errno);
            return false;
        }

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_URL, sanitisedUrl.c_str());
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, file);
        curl_easy_setopt(curl, CURLOPT_USERNAME, "KqO7FfIYaFQMQw");
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, ImageDownloadedCallback);
        curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1);
        curl_easy_setopt(curl, CURLOPT_USERAGENT, "Headlines/1.0");
        curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        struct curl_slist* hs = NULL;
        for (const std::string& header : headerValues)
        {
            hs = curl_slist_append(hs, header.c_str());
        }
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, hs);

        CURLcode rc = curl_easy_perform(curl);
        if (rc)
        {
            if (s_ErrorCodeMap.find(rc) != s_ErrorCodeMap.end())
            {
                g_warning("Httplient::DownloadImage error: %s - %s", sanitisedUrl.c_str(), s_ErrorCodeMap[rc].c_str());
            }
            else
            {
                g_warning("Httplient::DownloadImage error: %s - %d", sanitisedUrl.c_str(), (int)rc);
            }
            return false;
        }

        curl_easy_cleanup(curl);

        fclose(file);

        return true;
    }

    std::string HttpClient::Post(std::string url, std::string data, std::vector<std::string> headerValues) const
    {
        std::string response;
        ResponseCallbackArgs args{response};

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_easy_setopt(curl, CURLOPT_USERNAME, "KqO7FfIYaFQMQw");
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, GetResponseCallback);
        curl_easy_setopt(curl, CURLOPT_POST, 1L);
        curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, data.size());
        curl_easy_setopt(curl, CURLOPT_POSTFIELDS, data.c_str());
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*) &args);
        curl_easy_setopt(curl, CURLOPT_USERAGENT, "Headlines/1.0");
        struct curl_slist* hs = NULL;
        for (const std::string& header : headerValues)
        {
            hs = curl_slist_append(hs, header.c_str());
        }

        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, hs);

        CURLcode res = curl_easy_perform(curl);
        curl_easy_cleanup(curl);
        curl_slist_free_all(hs);

        if (AppSettings::Get()->GetBool("print_api_responses", false))
        {
            std::cout << "HttpClient::Post: " << url << std::endl;
            std::cout << "Response: " << util::Helpers::UnescapeHtml(response) << std::endl << std::endl;
        }

        if (res)
        {
            if (s_ErrorCodeMap.find(res) != s_ErrorCodeMap.end())
            {
                g_warning("Httplient::Post error: %s - %s", url.c_str(), s_ErrorCodeMap[res].c_str());
            }
            else
            {
                g_warning("Httplient::Post error: %s - %d", url.c_str(), (int)res);
            }
            return "";
        }

        return response;
    }

    std::string HttpClient::Post(std::string url, Json::Value argFields, std::vector<std::string> headerValues) const
    {
        std::string response;
        ResponseCallbackArgs args{response};

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_easy_setopt(curl, CURLOPT_USERNAME, "KqO7FfIYaFQMQw");
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, GetResponseCallback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*) &args);
        curl_easy_setopt(curl, CURLOPT_USERAGENT, "Headlines/1.0");
        curl_easy_setopt(curl, CURLOPT_ENCODING, "application/x-www-form-urlencoded");

        struct curl_slist* hs = NULL;
        for (const std::string& header : headerValues)
        {
            hs = curl_slist_append(hs, header.c_str());
        }

        curl_mime* form = NULL;
        curl_mimepart* field = NULL;
        form = curl_mime_init(curl);
        for (unsigned int i = 0; i < argFields.size(); ++i)
        {
            field = curl_mime_addpart(form);
            curl_mime_name(field, argFields[i]["name"].asCString());
            curl_mime_data(field, argFields[i]["value"].asCString(), CURL_ZERO_TERMINATED);
        }

        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, hs);
        curl_easy_setopt(curl, CURLOPT_MIMEPOST, form);

        CURLcode res = curl_easy_perform(curl);
        curl_easy_cleanup(curl);
        curl_slist_free_all(hs);

        if (AppSettings::Get()->GetBool("print_api_responses", false))
        {
            std::cout << "HttpClient::Post: " << url << std::endl;
            std::cout << "Response: " << util::Helpers::UnescapeHtml(response) << std::endl << std::endl;
        }

        if (res)
        {
            if (s_ErrorCodeMap.find(res) != s_ErrorCodeMap.end())
            {
                g_warning("Httplient::Post error: %s - %s", url.c_str(), s_ErrorCodeMap[res].c_str());
            }
            else
            {
                g_warning("Httplient::Post error: %s - %d", url.c_str(), (int)res);
            }
            return "";
        }

        return response;
    }

    std::string HttpClient::Put(std::string url, Json::Value argFields, std::vector<std::string> headerValues) const
    {
        std::string response;
        ResponseCallbackArgs args{response};

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "PUT"); /* !!! */
        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, GetResponseCallback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*) &args);
        curl_easy_setopt(curl, CURLOPT_USERAGENT, "Headlines/1.0");
        curl_easy_setopt(curl, CURLOPT_ENCODING, "application/x-www-form-urlencoded");

        curl_mime* form = NULL;
        curl_mimepart* field = NULL;
        /* Create the form */
        form = curl_mime_init(curl);

        for (unsigned int i = 0; i < argFields.size(); ++i)
        {
            field = curl_mime_addpart(form);
            ASSERT(argFields[i]["name"].isString(), "All put args must have a string name.");
            curl_mime_name(field, argFields[i]["name"].asCString());

            if (argFields[i]["value"].isString())
            {
                curl_mime_data(field, argFields[i]["value"].asCString(), CURL_ZERO_TERMINATED);
            }
            else
            {
                std::stringstream ss;
                Json::StreamWriterBuilder builder;
                Json::StreamWriter* writer = builder.newStreamWriter();
                writer->write(argFields[i]["value"], &ss);
                curl_mime_data(field, ss.str().c_str(), CURL_ZERO_TERMINATED);
            }
        }

        struct curl_slist* hs = NULL;
        for (const std::string& header : headerValues)
        {
            hs = curl_slist_append(hs, header.c_str());
        }
        hs = curl_slist_append(hs, "Expect:");

        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, hs);
        curl_easy_setopt(curl, CURLOPT_MIMEPOST, form);

        CURLcode res = curl_easy_perform(curl);
        curl_easy_cleanup(curl);

        if (AppSettings::Get()->GetBool("print_api_responses", false))
        {
            std::cout << "HttpClient::Get: " << url << std::endl;
            std::cout << "Response: " << util::Helpers::UnescapeHtml(response) << std::endl << std::endl;
        }

        if (res)
        {
            if (s_ErrorCodeMap.find(res) != s_ErrorCodeMap.end())
            {
                g_warning("Httplient::UploadMedia error: %s - %s", url.c_str(), s_ErrorCodeMap[res].c_str());
            }
            else
            {
                g_warning("Httplient::UploadMedia error: %s - %d", url.c_str(), (int)res);
            }
            return "";
        }

        return response;
    }

    std::string HttpClient::UploadMedia(std::string url, Json::Value argFields, const std::string& fileName, const std::string mimeType) const
    {
        std::string response;
        ResponseCallbackArgs args{response};

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, GetResponseCallback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*) &args);
        curl_easy_setopt(curl, CURLOPT_USERAGENT, "Headlines/1.0");

        curl_mime* form = NULL;
        curl_mimepart* field = NULL;
        struct curl_slist* headerlist = NULL;
        static const char buf[] = "Expect:";
        /* Create the form */
        form = curl_mime_init(curl);

        for (unsigned int i = 0; i < argFields.size(); ++i)
        {
            field = curl_mime_addpart(form);
            curl_mime_name(field, argFields[i]["name"].asCString());
            curl_mime_data(field, argFields[i]["value"].asCString(), CURL_ZERO_TERMINATED);
        }

        field = curl_mime_addpart(form);
        curl_mime_name(field, "file");
        curl_mime_filedata(field, fileName.c_str());

        field = curl_mime_addpart(form);
        curl_mime_name(field, "filename");
        curl_mime_data(field, fileName.c_str(), CURL_ZERO_TERMINATED);

        field = curl_mime_addpart(form);
        curl_mime_type(field, mimeType.c_str());

        headerlist = curl_slist_append(headerlist, buf);
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headerlist);
        curl_easy_setopt(curl, CURLOPT_MIMEPOST, form);

        CURLcode res = curl_easy_perform(curl);
        curl_easy_cleanup(curl);

        if (AppSettings::Get()->GetBool("print_api_responses", false))
        {
            std::cout << "HttpClient::Get: " << url << std::endl;
            std::cout << "Response: " << util::Helpers::UnescapeHtml(response) << std::endl << std::endl;
        }

        if (res)
        {
            if (s_ErrorCodeMap.find(res) != s_ErrorCodeMap.end())
            {
                g_warning("Httplient::UploadMedia error: %s - %s", url.c_str(), s_ErrorCodeMap[res].c_str());
            }
            else
            {
                g_warning("Httplient::UploadMedia error: %s - %d", url.c_str(), (int)res);
            }
            return "";
        }

        return response;
    }

    std::string HttpClient::Delete(std::string url, std::vector<std::string> headerValues) const
    {
        std::string response;
        ResponseCallbackArgs args{response};

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "DELETE"); /* !!! */
        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, GetResponseCallback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*) &args);
        curl_easy_setopt(curl, CURLOPT_USERAGENT, "Headlines/1.0");
        struct curl_slist* hs = NULL;
        for (const std::string& header : headerValues)
        {
            hs = curl_slist_append(hs, header.c_str());
        }

        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, hs);

        CURLcode res = curl_easy_perform(curl);
        curl_slist_free_all(hs);
        curl_easy_cleanup(curl);

        if (AppSettings::Get()->GetBool("print_api_responses", false))
        {
            std::cout << "HttpClient::Get: " << url << std::endl;
            std::cout << "Response: " << util::Helpers::UnescapeHtml(response) << std::endl << std::endl;
        }

        if (res)
        {
            if (s_ErrorCodeMap.find(res) != s_ErrorCodeMap.end())
            {
                g_warning("Httplient::Get error: %s - %s", url.c_str(), s_ErrorCodeMap[res].c_str());
            } else
            {
                g_warning("Httplient::Get error: %s - %d", url.c_str(), (int) res);
            }
            return "";
        }

        return response;
    }
}
